/*
 * microfsm.h
 *
 * Copyright (C) 2020 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
*/


#ifndef _MICROFSM_H_
#define _MICROFSM_H_

#include <stdint.h>


#ifndef NULL
#define NULL 0L
#endif


enum microfsm_rc_enum {
	MICROFSM_RC_SUCCESS	= 0,
	MICROFSM_RC_INVALIDARG,
};


enum microfsm_basesignal_enum {
	MICROFSM_SIG_EMPTY	= 0,
	MICROFSM_SIG_ENTRY,
	MICROFSM_SIG_EXIT,
	MICROFSM_SIG_USER,	// Initial user-defined signal id
};


typedef uint8_t microfsm_result_t;
typedef uint8_t microfsm_signal_t;


// Applications use this as the base (super) for new event types
typedef struct microfsm_event_s microfsm_event_t;
struct microfsm_event_s {
	microfsm_signal_t	signal;
};


// Applications create an instance of this for each FSM
typedef struct microfsm_s microfsm_t;



typedef microfsm_result_t (*microfsm_handler_t)(microfsm_t *me, microfsm_event_t *event);

struct microfsm_s {
        microfsm_handler_t	initial;     // initial state handler
        microfsm_handler_t	state;       // current state handler
	void			*userdata;   // opaque user data
};



#define MICROFSM_RC_HANDLED		((microfsm_result_t)0)
#define MICROFSM_RC_IGNORED		((microfsm_result_t)1)
#define MICROFSM_RC_TRANSITION		((microfsm_result_t)2)

#define MICROFSM_RET_HANDLED()		(MICROFSM_RC_HANDLED)
#define MICROFSM_RET_IGNORED()		(MICROFSM_RC_IGNORED)
#define MICROFSM_RET_TRANSITION(target)	\
	(((microfsm_t*)me)->state=(microfsm_handler_t)(target) , MICROFSM_RC_TRANSITION)

#define MICROFSM_GET_USERDATA(me)	(((microfsm_t*)me)->userdata)


// API
microfsm_result_t microfsm_init(microfsm_t *me, microfsm_handler_t initial);
microfsm_result_t microfsm_dispatch(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t microfsm_set_userdata(microfsm_t *me, void *userdata);


#endif


