/************************************************************************/
/* mailfilterd.c								*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cutil/cutil.h>

#include <mailfilterd.h>
#include <helpers.h>
#include <tcp.h>
#include <udp.h>
#include <pnetbuffer.h>
//#include <protocols.h>
#include <pfm.h>
#include <ioloop.h>
#include <logging.h>


#define LOG_PATH_DEFAULT	"/var/log/mailfilterd.log"

#define PATH_SIZE		1024

#define SYSLOG_BUFFER_SIZE	1024
#define SYSLOG_FACILITY		(20<<3)

#define LINEBUFFER_SIZE		2048
typedef struct _linebuffer_t linebuffer_t;
struct _linebuffer_t {
	char		buffer[LINEBUFFER_SIZE];
	int		len;
	int		read_oset;
};



// Globals
mailfilterd_t mailfilterd;


int process_command(int fd, int events, connection_data_t *cd, void *arg);
int log_entry(int log_level, char *format, ...);


int process_frontend_packet(int fd, int events, connection_data_t *cd, void *arg) {
	int		rc;
//	uint16_t	plen;
//	uint16_t	port;
//	int		read_len;
//	int		vote;
//	socklen_t	saddr_len;
	char		buffer_in[PACKET_BUFFER_SIZE];
	char		buffer_out[PACKET_BUFFER_SIZE];
//	char		*p;
	pnetbuffer_t	*nb_in;
	pnetbuffer_t	*nb_out;
	io_context_t	ctx;

	//************************************
	// init
	//************************************
//	vote =  VOTE_PASS;
	ctx.dest_port = cd->port;
	ctx.src_fd = fd;

	log_entry(
		LOG_LEVEL_NOTICE,
		"process_frontend_packet()");

	//************************************
	// read packet
	//************************************
	nb_in = pnetbuffer_init(buffer_in, PACKET_BUFFER_SIZE);
//	pnetbuffer_get_writable(nb_in, &p, &plen);
//	saddr_len = sizeof(struct sockaddr);
//	read_len = recvfrom(fd, p, (int)plen, 0, &ctx.src_addr, &saddr_len);
//	if (read_len == -1) {
//		rc = errno;
//		return rc;
//	}
//	rc = pnetbuffer_set_written(nb_in, (uint16_t)read_len);
	
//	//************************************
//	// input packet validations
//	//************************************
//	if (ctx.src_addr.sa_family != AF_INET) {
//		log_entry(
//			LOG_LEVEL_WARNING,
//			"Non-IP packet received, dropping");
//		pnetbuffer_reset(nb_in);
//		return 0;
//	}

	//************************************
	// process input packet
	//************************************
	nb_out = pnetbuffer_init(buffer_out, PACKET_BUFFER_SIZE);
	rc = pfm_process_packet(mailfilterd.pfm_manager, &ctx, nb_in, nb_out);
	log_entry(
		LOG_LEVEL_WARNING,
		"pfm_process_packet() vote=%d", rc);

	(void)rc; // silence compiler warning
	return 0;
}



/*
int protocol_add_filter(compare_t *filter)
{
	return slist_item_add(&mailfilterd.proto_filter_list[filter->protocol], filter);
}


int protocol_remove_filter(compare_t *filter)
{
	return slist_item_remove(&mailfilterd.proto_filter_list[filter->protocol], filter);
}


int protocol_count_filter(uint16_t protocol)
{
	int	c;
	int	rc;
	rc = slist_item_count(mailfilterd.proto_filter_list[protocol], &c);
	(void)rc;
	return c;
}
*/



int command_server_callback(int fd, int events, connection_data_t *cd, void *arg)
{
	int	     rc;
	int	     c_flags;
	int	     c_fd;
	socklen_t       c_len;
	struct sockaddr c_sa;

	c_len = sizeof(struct sockaddr);
	c_fd = accept(fd, &c_sa, &c_len);
	if (c_fd == -1) {
		rc = errno;
		log_entry(LOG_LEVEL_WARNING,
			"Failed to accept new admin client connection"
			" rc=%d listen_fd=%d",
			rc, fd);
		return rc;
	}
	c_flags = fcntl(c_fd, F_GETFL, 0);
	rc = fcntl(c_fd, F_SETFL, c_flags | O_NONBLOCK);
	if (rc == -1) {
		rc = errno;
		log_entry(LOG_LEVEL_WARNING,
			"Failed to set O_NONBLOCK on new client"
			" connection rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		return rc;
	}

//	client = gclient_new(c_fd, gserver.log);
//	if (!client) {
//		rc = errno;
//		log_entry(LOG_LEVEL_ERROR,
//			"Failed to allocate new client structure rc=%d",
//			rc);
//		c_flags = tcp_shutdown(c_fd);
//		return rc;
//	}

	rc = ioloop_register_fd(mailfilterd.io_loop,
				c_fd,
				CD_TYPE_COMMAND,
				mailfilterd.command_port,
				process_command,
				(void*)c_fd);
	if (rc) {
		log_entry(LOG_LEVEL_ERROR,
			"Failed to register new command client FD with ioloop"
			" subsystem rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		return rc;
	}
	log_entry(LOG_LEVEL_NOTICE,
		"fd=%d Accepted new command client connection",
		c_fd);

	return 0;
}




int process_command(int fd, int events, connection_data_t *cd, void *arg) {
	int		rc;
//	uint16_t	port;
	uint16_t	plen;
	int		read_len;
	int		res_len;
	int		i;
//	uint8_t		u8;
//	uint16_t	u16;
	struct sockaddr	saddr;
	socklen_t	saddr_len;
	char		cmdline[LINEBUFFER_SIZE];
	char		buffer[PACKET_BUFFER_SIZE];
	char		res_buffer[PACKET_BUFFER_SIZE];
	char		*argv[64];
	char		*p;
	char		*sp;
	pnetbuffer_t	*req_nb;
	pnetbuffer_t	*res_nb;
//	SList		*item;
//	pfm_module_t	*module;

	//************************************
	// init
	//************************************
//	u16 = 0;
	res_len = 0;
	req_nb = pnetbuffer_init(buffer, PACKET_BUFFER_SIZE);
	res_nb = pnetbuffer_init(res_buffer, PACKET_BUFFER_SIZE);

	log_entry(
		LOG_LEVEL_DEBUG,
		"process_command() entry");

	//************************************
	// read packet
	//************************************
	pnetbuffer_get_writable(req_nb, &p, &plen);
	saddr_len = sizeof(struct sockaddr);
	read_len = recvfrom(fd, p, (int)plen, 0, &saddr, &saddr_len);
	if (read_len == -1) {
		// an error occurred during a socket read attempt
		rc = errno;
		log_entry(
			LOG_LEVEL_ERROR,
			"recv() rc=%d closing client connection",
			rc);
		rc = ioloop_unregister_fd(mailfilterd.io_loop, fd);
		rc = tcp_shutdown(fd);
		return rc;
	}
	if (read_len == 0) {
		// the client closed their end of the connection
		rc = errno;
		log_entry(
			LOG_LEVEL_NOTICE,
			"recv() rc=%d client disconnected closing connection",
			read_len);
		rc = ioloop_unregister_fd(mailfilterd.io_loop, fd);
		rc = tcp_shutdown(fd);
		return rc;
	}

//	log_entry(
//		LOG_LEVEL_NOTICE,
//		"recv() read_len=%d",
//		read_len);
	rc = pnetbuffer_set_written(req_nb, (uint16_t)read_len);
	if (read_len < 2) goto no_command_response;

	//************************************
	// process command header
	//************************************
	rc = pnetbuffer_read_data(req_nb, cmdline, read_len);
	if (rc) {
		log_entry(
			LOG_LEVEL_WARNING,
			"pnetbuffer_read_string() rc=%d",
			rc);
		return rc;
	}
	if (!strncmp(cmdline, "ping", 4)) {
		rc = pnetbuffer_add_string(res_nb, "0 pong\n");

	} else if (!strncmp(cmdline, "quit", 4)) {
		log_entry(
			LOG_LEVEL_NOTICE,
			"client requested to close connection",
			read_len);
		rc = ioloop_unregister_fd(mailfilterd.io_loop, fd);
		rc = tcp_shutdown(fd);
		return rc;

	} else if (!strncmp(cmdline, "pfm command ", 12)) {
		pnetbuffer_reset(req_nb);
		p = &cmdline[12];
		// unused string counter
		//rc = pnetbuffer_add_uint8(req_nb, 2);
		// pfm module name
		p = strtok_r(p, " ", &sp);
		rc = pnetbuffer_add_strings(req_nb,
					&p,
					1,
					PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		// pfm command
		i = 0;
		while (1) {
			p = strtok_r(NULL, " \r\n", &sp);
			if (!p) break;
			argv[i] = p;
			i++;
			if (i == 64) break; // TODO command too long
		}
		rc = pnetbuffer_add_strings(req_nb,
					argv,
					i,
					PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		if (rc) {
			log_entry(
				LOG_LEVEL_WARNING,
				"Failed to parse pfm command to module rc=%d",
				rc);
			rc = pnetbuffer_add_string(res_nb,
				"2 Failed to parse pfm command to module\n");
			goto process_command_response;
		}
		rc = pfm_process_command(mailfilterd.pfm_manager,
			req_nb,
			res_nb);

	} else {
		rc = pnetbuffer_add_string(res_nb, "1 invalid command\n");

	}

/*
	switch (u16) { // command
		case CMD_PING:
			rc = pnetbuffer_add_uint16(res_nb, CMD_PING|CMD_RESPONSE);
			break;
		case CMD_SET_LOG_LEVEL:
			rc = pnetbuffer_read_uint8(req_nb, &u8);
			if (rc)  {
				rc = pnetbuffer_add_uint16(res_nb, CMD_ERROR);
				break;
			}
			logging_set_log_level(mailfilterd.log, u8);
			rc = pnetbuffer_add_uint16(res_nb, CMD_SET_LOG_LEVEL|CMD_RESPONSE);
			log_entry(
				LOG_LEVEL_ERROR,
				"log level set to %s",
				logging_level_to_string(u8));
			break;
		case CMD_LIST_PFMS:
			rc = pnetbuffer_add_uint16(res_nb, CMD_LIST_PFMS|CMD_RESPONSE);
			// module count
			u16 = 0;
			item = pfm_module_list(mailfilterd.pfm_manager);
			while (item) {
				item = slist_get_next(item);
				u16++;
			}
			rc = pnetbuffer_add_uint16(res_nb, u16);
			
			// module info list
			item = pfm_module_list(mailfilterd.pfm_manager);
			while (item) {
				module = slist_get_data(item);
				u8 = strlen(module->info->name);
				rc = pnetbuffer_add_uint8(res_nb, u8); // name len
				rc = pnetbuffer_add_data(res_nb, module->info->name, u8); // name
				rc = pnetbuffer_add_uint16(res_nb, module->info->version); // version
				item = slist_get_next(item);
			}
			break;
		case CMD_PFM_COMMAND:
			rc = pnetbuffer_add_uint16(res_nb, CMD_PFM_COMMAND|CMD_RESPONSE);
			rc = pfm_process_command(mailfilterd.pfm_manager, req_nb, res_nb);
			break;

		default:
			rc = pnetbuffer_add_uint16(res_nb, CMD_ERROR);
			break;
	}
*/

process_command_response:
	pnetbuffer_get_readable(res_nb, &p, &plen);
	rc = sendto(fd, p, (int)plen, 0, &saddr, saddr_len);
	if (rc == -1) {
		rc = errno;
		log_entry(
			LOG_LEVEL_ERROR,
			"send() rc=%d",
			res_len);
		return rc;
	}

no_command_response:
	return 0;
}


int log_entry_va(int log_level, char *format, va_list args)
{
	va_list		va;
	int		rc;
	int		pri;
	int		len1;
	int		len2;
	char		*p;
	char		sstr[SYSLOG_BUFFER_SIZE];
	time_t		t;
	struct tm	*lt = NULL;

	if (log_level > mailfilterd.log->level)
		return 0; // dont log above current level

	// handle stdio/logfile logging
	va_copy(va, args);
	logging_entry_va(mailfilterd.log,
			log_level,
			format,
			va);
	va_end(va);

	// handle syslog (if enabled)
	if (mailfilterd.syslog_fd != -1) {
		t = time(NULL);
		lt = localtime(&t);
		pri = SYSLOG_FACILITY + log_level;
		len1 = snprintf(sstr,
			SYSLOG_BUFFER_SIZE,
			"<%d>1 %04d-%02d-%02dT%02d:%02d:%02d %s mailfilterd[%d] - ",
			pri,
			lt->tm_year+1900,
			lt->tm_mon+1,
			lt->tm_mday,
			lt->tm_hour,
			lt->tm_min,
			lt->tm_sec,
			mailfilterd.hostname,
			mailfilterd.pid);
		p = &sstr[len1];
		va_copy(va, args);
		len2 = vsnprintf(p,
			SYSLOG_BUFFER_SIZE-len1,
			format,
			va);
		va_end(va);
		len1 += len2;
		if (len1 < SYSLOG_BUFFER_SIZE) {
			rc = send(mailfilterd.syslog_fd, sstr, len1, 0);
			if (rc == -1) {
				rc = errno;
				if (rc == EAGAIN) return 0; // one-try best effort
				if (rc == ECONNREFUSED) {
					// the remote syslog service is down
					// TODO: disable and backoff for a while?
					printf("ERROR The remote syslog service is down.\n");
					return 0;
				}
				// TODO: disable and backoff for a while?
				udp_shutdown(mailfilterd.syslog_fd);
				mailfilterd.syslog_fd = -1;
				printf("ERROR failure while sending to syslog server rc=%d."
					" Further use has been disabled.\n",
					errno);
			}
		}
	}
	return 0;
}


int log_entry(int log_level, char *format, ...)
{
	va_list		va;

	va_start(va, format);
	log_entry_va(
			log_level,
			format,
			va);
	va_end(va);
	return 0;
}


int str_to_array(char *str, char **array, int maxlen, int maxcount, int *count)
{
	int		i;
	char		*p;
	char		*tok;

	i = 0;
	p = strtok_r(str, " ", &tok);
	while (p) {
		strncpy(array[i], p, maxlen-1);
		array[i][maxlen-1] = 0;
		i++;
		p = strtok_r(NULL, " ", &tok);
	}
	*count = i;
	return 0;
}


int linebuffer_fill(linebuffer_t *lb, int fd)
{
	size_t		len;
	size_t		c;

	c = LINEBUFFER_SIZE - lb->len;
	len = read(fd, &lb->buffer[lb->len], c);
//	printf("%s() read() rc=%d\n", __func__, len);
	if (len == -1)
		return -1;
	lb->len += len;
//	lb->buffer[lb->len] = 0;
	return len;
}


char *linebuffer_get_line(linebuffer_t *lb, int fd)
{
	int		i;
	int		so;
	int		lr;

	i = lb->read_oset;
	so = i;
	while (1) {
//		printf("%s() loop i=%d lb->len=%d\n", __func__, i, lb->len);
		if (i >= lb->len) {
			if (i == LINEBUFFER_SIZE) {
				// slide remaining bytes first
				strncpy(lb->buffer, &lb->buffer[so], i-so);
				lb->read_oset = 0;
				lb->len = i-so;
				i = lb->read_oset;
				so = i;
			}
			lr = linebuffer_fill(lb, fd);
			if (lr == -1)
				return NULL;
			if (lr == 0) {
				break;
			}
		}
		if (lb->buffer[i] == '\n') {
			lb->buffer[i] = 0;
			lb->read_oset = i+1;
			return &lb->buffer[so];
		}
		i++;
	}
	// EOF reached
	if ((lb->len-lb->read_oset) != 0) {
		// last line missing NL
		lb->read_oset = 0;
		lb->len = 0;
		return &lb->buffer[so];
	}
	return NULL;
}


int syslog_connect(char *hostname, char *port)
{
	int		rc;
	rc = udp_connect(mailfilterd.syslog_hostname,
			mailfilterd.syslog_port);
	if (rc == -1) {
		rc = errno;
		return rc;
	}
	mailfilterd.syslog_fd = rc;
	return 0;
}


int config_load(char *filepath, int options)
{
	int		rc;
	int		i;
	int		ln;
	int		fd;
	int		count;
	uint16_t	u16;
	char		*p;
	linebuffer_t	lb;
	char		*array[PFM_COMMAND_TOKEN_MAX];
	char		path[PATH_SIZE];
	char		cmd_nb_buffer[2048];
	pnetbuffer_t	*cmd_nb;
	char		res_nb_buffer[2048];
	pnetbuffer_t	*res_nb;

	rc = 0;

	cmd_nb = pnetbuffer_init(cmd_nb_buffer, 2048);
	res_nb = pnetbuffer_init(res_nb_buffer, 2048);

	for (i=0 ; i<PFM_COMMAND_TOKEN_MAX ; i++)
		array[i] = malloc(PFM_COMMAND_STR_SIZE);

	fd = open(filepath, O_RDONLY);
	if (fd == -1)
		return errno;
	lb.len = 0;
	lb.read_oset = 0;

	ln = 0;
	while (1) {
		p = linebuffer_get_line(&lb, fd);
		if (!p)
			break;
		ln++;
//		log_entry(
//			LOG_LEVEL_INFO,
//			"config line(%d)='%s'",
//			ln,
//			p);
		if (p[0] == '#') continue; // skip comment lines
		if (p[0] == '\n') continue; // skip empty lines
		if (p[0] == 0) continue; // skip empty string
		rc = str_to_array(p, array, PFM_COMMAND_STR_SIZE, PFM_COMMAND_TOKEN_MAX, &count);
		if (rc) {
			log_entry(
				LOG_LEVEL_ERROR,
				"Failed to parse config file line %d",
				ln);
			goto config_load_exit;
		}

		i = 0;
		if (!strcmp(array[i], "set")) {
			i++;
			if (!strcmp(array[i], "log_file")) {
				i++;
				mailfilterd.log_path = strdup(array[i]);
				rc = logging_reopen(mailfilterd.log, mailfilterd.log_path, 0);
				if (rc) {
					log_entry(
						LOG_LEVEL_ERROR,
						"Failed to open log file %s rc=%d config line %d",
						mailfilterd.log_path,
						rc,
						ln);
					goto config_load_exit;
				}
			} else if (!strcmp(array[i], "log_level")) {
				i++;
				logging_set_log_level(mailfilterd.log,
					logging_string_to_level(array[i]));
			} else if (!strcmp(array[i], "command_bind_address")) {
				i++;
				mailfilterd.command_bind_address = strdup(array[i]);
				log_entry(
					LOG_LEVEL_NOTICE,
					"Command bind address set to %s",
					mailfilterd.command_bind_address);
			} else if (!strcmp(array[i], "command_bind_port")) {
				i++;
				mailfilterd.command_bind_port = strdup(array[i]);
				log_entry(
					LOG_LEVEL_NOTICE,
					"Command bind port set to %s",
					mailfilterd.command_bind_port);
			} else if (!strcmp(array[i], "bind_address")) {
				i++;
				mailfilterd.bind_hostname = strdup(array[i]);
				log_entry(
					LOG_LEVEL_NOTICE,
					"Service bind address set to %s",
					mailfilterd.bind_hostname);
			} else if (!strcmp(array[i], "syslog_destination")) {
				i++;
				mailfilterd.syslog_hostname = strdup(array[i]);
				i++;
				if (array[i]) {
					mailfilterd.syslog_port = strdup(array[i]);
				} else {
					mailfilterd.syslog_port = strdup("514");
				}
				rc = syslog_connect(mailfilterd.syslog_hostname,
					mailfilterd.syslog_port);
				if (rc) {
					log_entry(
						LOG_LEVEL_WARNING,
						"Syslog destination connect failed to %s:%s rc=%d",
						mailfilterd.syslog_hostname,
						mailfilterd.syslog_port,
						rc);
					goto config_load_exit;
				}
				log_entry(
					LOG_LEVEL_NOTICE,
					"Syslog destination set to %s:%s",
					mailfilterd.syslog_hostname,
					mailfilterd.syslog_port);
			} else {
				log_entry(
					LOG_LEVEL_ERROR,
					"Invalid config line %d",
					ln);
				rc = EINVAL;
				goto config_load_exit;
			}
		} else if (!strcmp(array[i], "pfm")) {
			i++;
			if (!strcmp(array[i], "command")) {
				i++;
				// pfm command ...
				pnetbuffer_reset(cmd_nb);
				pnetbuffer_reset(res_nb);
				rc = pnetbuffer_add_strings(cmd_nb,
						&array[i],
						1,
						PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
				i++;
				rc = pnetbuffer_add_strings(cmd_nb,
						&array[i],
						count-i,
						PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
				rc = pfm_process_command(mailfilterd.pfm_manager,
						cmd_nb,
						res_nb);
				rc = pnetbuffer_read_uint16(res_nb, &u16);
				if (rc || u16 != PFM_ERR_SUCCESS) {
					log_entry(
						LOG_LEVEL_ERROR,
						"pfm command failed on line %d",
						ln);
					goto config_load_exit;
				}
			} else if (!strcmp(array[i], "load")) {
				i++;
				// pfm load modulename
				rc = snprintf(path, PATH_SIZE, "%s/pfm_%s.so", PKGLIBDIR, array[i]);
				if (rc >= PATH_SIZE) {
					log_entry(
						LOG_LEVEL_ERROR,
						"PFM module path length would exceed buffer size");
					return 2;
				}
				rc = pfm_module_load(mailfilterd.pfm_manager, path);
				if (rc != 0) {
					(void)pfm_close(mailfilterd.pfm_manager);
					log_entry(
						LOG_LEVEL_ERROR,
						"Unable to load pfm module");
					return 2;
				}
				log_entry(LOG_LEVEL_NOTICE,
					"Successfully loaded %s pfm module",
					array[i]);
			}
		} else {
			log_entry(
				LOG_LEVEL_ERROR,
				"Invalid config line %d",
				ln);
			rc = EINVAL;
			goto config_load_exit;
		}
	}
config_load_exit:
	(void)close(fd);
	return rc;
}


int main (int argc, char **argv, char**env)
{
	int		i;
	int		rc;
//	int		fd;
//	uint16_t	port;
//	char		portname[8];
	char		*p;

	memset(&mailfilterd, 0, sizeof(mailfilterd));
	mailfilterd.poll_timeout = POLL_TIMEOUT_DEFAULT * 1000;
	mailfilterd.dest_hostname = NULL;
	//mailfilterd.bind_hostname = "0.0.0.0";
	mailfilterd.config_path = SYSCONFDIR "/mailfilterd.conf";
	mailfilterd.bind_hostname = "127.0.0.1";
	mailfilterd.command_bind_address = "127.0.0.1";
	mailfilterd.command_bind_port = "5200";
	mailfilterd.log_path = NULL; // stdout logging by default
	mailfilterd.syslog_fd = -1; // syslog disabled by default
	// collect pid/short hostname for syslog messages
	mailfilterd.pid = getpid();
	rc = gethostname(mailfilterd.hostname, HOSTNAME_SIZE);
	if (rc != 0) {
		strcpy(mailfilterd.hostname, "localhost");
	}
	p = strstr(mailfilterd.hostname, ".");
	if (p) *p = 0; // trim at first period

	if (argc > 1) {
		i = 1;
		while (i < argc) {
			if (!strcmp(argv[i], "-b")) {
				i++;
				if (i >= argc) goto main_exit_usage;
				mailfilterd.bind_hostname = argv[i];
			} else if (!strcmp(argv[i], "-c")) {
				i++;
				if (i >= argc) goto main_exit_usage;
				mailfilterd.config_path = argv[i];
			} else if (!strcmp(argv[i], "-l")) {
				i++;
				if (i >= argc) goto main_exit_usage;
				mailfilterd.log_path = argv[i];
			} else {
				goto main_exit_usage;
			}
			i++;
		}
	}
	if (mailfilterd.bind_hostname == NULL) goto main_exit_usage;

	mailfilterd.log = logging_open(mailfilterd.log_path, 0);
	if(!mailfilterd.log) {
		printf("ERROR Unable to initialize logging subsystem\n");
		return 2;
	}

	mailfilterd.io_loop = ioloop_open(0);
	if(!mailfilterd.io_loop) {
		log_entry(
			LOG_LEVEL_ERROR,
			"Unable to initialize ioloop subsystem");
		return 2;
	}
	log_entry(
		LOG_LEVEL_NOTICE,
		"Successfully initialized ioloop subsystem");

	mailfilterd.pfm_init.ioloop = mailfilterd.io_loop;
	mailfilterd.pfm_manager = pfm_open(&mailfilterd.pfm_init);
	if(!mailfilterd.pfm_manager) {
		log_entry(
			LOG_LEVEL_ERROR,
			"Unable to initialize pfm subsystem");
		return 2;
	}
	rc = pfm_set_option(mailfilterd.pfm_manager,
			PFM_OPT_SET_LOG_CALLBACK,
			log_entry_va);
	log_entry(
		LOG_LEVEL_NOTICE,
		"Successfully initialized pfm subsystem");


	rc = config_load(mailfilterd.config_path, 0);
	if (rc != 0) {
		log_entry(
			LOG_LEVEL_ERROR,
			"Failure loading configuration file");
		return 2;
	}


	// command socket
	mailfilterd.command_port = atoi(mailfilterd.command_bind_port);
	mailfilterd.command_fd = tcp_bind(mailfilterd.command_bind_address,
					mailfilterd.command_bind_port);
	if (mailfilterd.command_fd == -1) {
		log_entry(
			LOG_LEVEL_ERROR,
			"Unable to bind command listener socket");
		return 2;
	}
	log_entry(
		LOG_LEVEL_NOTICE,
		"Bind command listener for port %d succeeded fd=%d",
		mailfilterd.command_port,
		mailfilterd.command_fd);
	rc = ioloop_register_fd(mailfilterd.io_loop,
				mailfilterd.command_fd,
				CD_TYPE_COMMAND,
				mailfilterd.command_port,
				command_server_callback,
				(void*)&mailfilterd.command_port);


/* moved into pfm modules
	// Bind Listeners
	for (port=1 ; port<65535 ; port++) {
		//rc = pfm_module_bind_module_to_port(mailfilterd.pfm_manager, module->info->name, 1053);
		if (mailfilterd.pfm_manager->port_hash[port] != NULL) {
			sprintf(portname, "%d", port);
			fd = tcp_bind(mailfilterd.bind_hostname, portname);
			if (fd == -1) {
				log_entry(
					LOG_LEVEL_ERROR,
					"Unable to bind listener socket for %s:%d",
					mailfilterd.bind_hostname,
					port);
				return 2;
			}

			rc = ioloop_register_fd(mailfilterd.io_loop,
						fd,
						CD_TYPE_FRONTEND,
						port,
						process_frontend_packet,
						NULL);
			log_entry(
				LOG_LEVEL_NOTICE,
				"Bind listener for port %d succeeded fd=%d",
				port,
				fd);
		}
	}
*/

	log_entry(
		LOG_LEVEL_NOTICE,
		"Entering main loop");
	rc = ioloop_run(mailfilterd.io_loop, 0, 5000);
	log_entry(
		LOG_LEVEL_WARNING,
		"The daemon has exited the IO loop rc=%d",
		rc);

	return rc;

main_exit_usage:
	printf ("Usage: mailfilterd [options]\n");
	printf ("Options:\n");
	printf (" -b hostname|IP   frontend bind address(default=localhost\n");
	printf (" -c configfile    path to config file(default=%s\n", SYSCONFDIR "/mailfilterd.conf");
	printf (" -l filepath      path to log file(default=stdout)\n");
	printf ("\n");
	return 2;
}



