/************************************************************************/
/* mailfilterd.c								*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifndef _UDPFILTERD_H_
#define _UDPFILTERD_H_


#include <stdint.h>
#include <cutil/cutil.h>

#include <helpers.h>
#include <tcp.h>
#include <pnetbuffer.h>
//#include <protocols.h>
#include <pfm.h>
#include <ioloop.h>
#include <logging.h>



#define POLL_TIMEOUT_DEFAULT		10
#define CONTROL_PORT			6181
#define PACKET_BUFFER_SIZE		16384
#define POLL_MAX_SOCKETS		16384
#define HOSTNAME_SIZE			64


typedef struct _ethertype_count_t ethertype_count_t;
struct _ethertype_count_t {
        uint32_t        arp_packets;
        uint32_t        ipv4_packets;
};


typedef struct _mailfilterd_t mailfilterd_t;
struct _mailfilterd_t {
	char			hostname[HOSTNAME_SIZE];	// for syslog use
	int			pid;				// for syslog use

	char			*command_bind_address;// command socket listen address
	char			*command_bind_port;// command socket listen port
	uint16_t		command_port;   // command socket port
	int			command_fd;     // local command socket

	int			poll_timeout;
	char			*config_path;	// config file path
	char			*dest_hostname; // destination hostname
						// for GENEVE packets
	char			*bind_hostname; // bind hostname for
						// GENEVE packet sourcing

//	control_state_t		control_state;

	ethertype_count_t	ethertype_counts;

	// one SList per protocol
	SList			*proto_filter_list[8100];

	pfm_init_t		pfm_init;
	pfm_manager_t		*pfm_manager;

	// I/O-related
	ioloop_t		*io_loop;


	// logging
	char			*log_path;
	logging_t		*log;

	// syslog target
	char			*syslog_hostname;
	char			*syslog_port;
	int			syslog_fd;
};



#endif


